import abc
import os
import uuid
from os import listdir
from os.path import isdir, isfile, join, basename
from shutil import rmtree
from typing import final


class DataHandle(metaclass=abc.ABCMeta):
    __baltic_data_prefix: final = 'BalticLSC-'
    __uuid_length: final = 6

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'download') and
                callable(subclass.download) and
                hasattr(subclass, 'upload') and
                callable(subclass.upload) and
                hasattr(subclass, 'check_connection') and
                callable(subclass.check_connection) or
                NotImplemented)

    @abc.abstractmethod
    def __init__(self, pin_name: str, pins_configuration: []):
        self._pin_configuration = next(pc for pc in pins_configuration if pc.pin_name == pin_name)
        self._local_path = os.getenv('LOCAL_TMP_PATH', '/balticLSC_tmp')
        self._create_local()

    @abc.abstractmethod
    def download(self, handle: {}) -> str:
        pass

    @abc.abstractmethod
    def upload(self, local_path: str) -> {}:
        pass

    @abc.abstractmethod
    def check_connection(self, handle: {}):
        pass

    def _create_local(self):
        if '.' in os.path.split(self._local_path)[-1]:
            file_path = os.path.dirname(self._local_path)
        else:
            file_path = self._local_path

        if isdir(file_path):
            return

        os.mkdir(file_path)

    def _clear_local(self):
        if isdir(self._local_path):
            rmtree(self._local_path, ignore_errors=True)
        elif isfile(self._local_path):
            os.remove(self._local_path)

    def _add_guids_to_file_names(self, local_path: str):
        for f in (f for f in listdir(local_path) if isfile(join(local_path, f))):
            file_name = basename(f)
            new_file_name =\
                self.__baltic_data_prefix + str(uuid.uuid4())[:self.__uuid_length] + '-'\
                + file_name if not file_name.startswith(self.__baltic_data_prefix) else file_name[len(
                    self.__baltic_data_prefix)+self.__uuid_length:]
            os.rename(f, join(new_file_name, local_path))


import logging
logging.basicConfig(
    level=logging.DEBUG, format='%(asctime)s %(levelname)-8s %(message)s (%(filename)s:%(lineno)d)',
    datefmt='%Y-%m-%d:%H:%M:%S')
logger = logging.getLogger('api')



import abc
import json
from http import HTTPStatus
from typing import Dict

from balticlsc.computation_module.baltic_lsc.data_handle import DataHandle
from balticlsc.computation_module.data_access.mongo_data_handle import MongoDBHandle
from balticlsc.computation_module.baltic_lsc.job_registry import JobRegistry
from balticlsc.computation_module.data_model.messages import Status
from balticlsc.computation_module.api_access.tokens_proxy import TokensProxy


class IDataHandler(metaclass=abc.ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'obtain_data_item') and
                callable(subclass.obtain_data_item) and
                hasattr(subclass, 'obtain_data_items') and
                callable(subclass.obtain_data_items) and
                hasattr(subclass, 'obtain_data_items_dim') and
                callable(subclass.obtain_data_items_dim) and
                hasattr(subclass, 'send_data_item') and
                callable(subclass.send_data_item) and
                hasattr(subclass, 'send_token') and
                callable(subclass.send_token) and
                hasattr(subclass, 'finish_processing') and
                callable(subclass.finish_processing) and
                hasattr(subclass, 'send_ack_token') and
                callable(subclass.send_ack_token) or
                NotImplemented)

    @abc.abstractmethod
    def obtain_data_item(self, pin_name: str) -> str or None:
        pass

    @abc.abstractmethod
    def obtain_data_items(self, pin_name: str) -> []:
        pass

    @abc.abstractmethod
    def obtain_data_items_dim(self, pin_name: str) -> ([], []):
        pass

    @abc.abstractmethod
    def send_data_item(self, pin_name: str, data: str, is_final: bool, msg_uid: str = None) -> int:
        pass

    @abc.abstractmethod
    def send_token(self, pin_name: str, values: str, is_final: bool, msg_uid: str = None) -> int:
        pass

    @abc.abstractmethod
    def finish_processing(self) -> int:
        pass

    @abc.abstractmethod
    def send_ack_token(self, msg_ids: [], is_final: bool) -> int:
        pass


class DataHandler(IDataHandler):

    def __init__(self, registry: JobRegistry, pins_configuration: []):
        self.__registry = registry
        self.__pins_configuration = pins_configuration
        self.__tokens_proxy = TokensProxy()
        self.__data_handles = {}

    def obtain_data_item(self, pin_name: str) -> str or None:
        (values, sizes) = self.obtain_data_items_dim(pin_name)
        if values is None or 0 == len(values):
            return None
        if sizes is None and 1 == len(values):
            return values[0]
        raise Exception('Improper call - more than one data item exists for the pin')

    def obtain_data_items(self, pin_name: str) -> []:
        (values, sizes) = self.obtain_data_items_dim(pin_name)
        if sizes is not None and 1 == len(sizes):
            return values
        raise Exception('Improper call - more than one dimension exists for the pin')

    def obtain_data_items_dim(self, pin_name: str) -> ([], []):
        (values, sizes) = self.__registry.get_pin_values_dim(pin_name)
        values_object = list(map(lambda v: None if v is None or not v else json.loads(v), values))
        d_handle = self.get_data_handle(pin_name)
        data_items = list(map(lambda vo: None if vo is None else d_handle.download(vo), values_object))
        return data_items, sizes

    def send_data_item(self, pin_name: str, data: str, is_final: bool, msg_uid: str = None) -> int:
        d_handle = self.get_data_handle(pin_name)
        new_handle = d_handle.upload(data)
        return self.send_token(pin_name, json.dumps(new_handle), is_final, msg_uid)

    def send_token(self, pin_name: str, values: str, is_final: bool, msg_uid: str = None) -> int:
        if msg_uid is None:
            msg_uid = self.__registry.get_base_msg_uid()

        return 0 if HTTPStatus.OK == self._DataHandler__tokens_proxy.send_output_token(pin_name, values, msg_uid, is_final) else -1

    def finish_processing(self) -> int:
        msg_ids = self.__registry.get_all_msg_uids()
        self.__registry.set_status(Status.COMPLETED)
        return self.send_ack_token(msg_ids, True)

    def fail_processing(self, note: str):
        msg_ids = self.__registry.get_all_msg_uids()
        self.__registry.set_status(Status.FAILED)
        if HTTPStatus.OK == self._DataHandler__tokens_proxy.send_ack_token(msg_ids, True, True, note):
            self.__registry.clear_messages(msg_ids)
            return 0
        return -1

    def send_ack_token(self, msg_ids: [], is_final: bool) -> int:
        if HTTPStatus.OK == self._DataHandler__tokens_proxy.send_ack_token(msg_ids, is_final):
            self.__registry.clear_messages(msg_ids)
            return 0

        return -1

    def check_connection(self, pin_name: str, handle: Dict[str, str] = None) -> int:
        try:
            d_handle = self.get_data_handle(pin_name)
            return d_handle.check_connection(handle)
        except ValueError:
            raise ValueError('Cannot check connection for a pin of type \"Direct\"')

    def get_data_handle(self, pin_name: str) -> DataHandle:
        if pin_name in self.__data_handles:
            return self.__data_handles[pin_name]
        access_type = self.__registry.get_pin_configuration(pin_name).access_type
        if 'Direct' == access_type:
            raise ValueError('Cannot create a data handle for a pin of type \"Direct\"')
        if 'MongoDB' == access_type:
            handle = MongoDBHandle(pin_name, self.__pins_configuration)
        else:
            raise NotImplementedError('AccessType (' + access_type + ') not supported by the DataHandler, has to '
                                                                         'be handled manually')
        self.__data_handles[pin_name] = handle
        return handle



from balticlsc.computation_module.api_access.job_controller import init_job_controller, TokenListener
from balticlsc.computation_module.baltic_lsc.data_handler import IDataHandler
from balticlsc.computation_module.baltic_lsc.job_registry import IJobRegistry


class MyTokenListener(TokenListener):

    def __init__(self, registry: IJobRegistry, data: IDataHandler):
        super().__init__(registry, data)

    def data_received(self, pin_name: str):
        # Place your code here:
        pass

    def optional_data_received(self, pin_name: str):
        # Place your code here:
        pass

    def data_ready(self):
        # Place your code here:
        pass

    def data_complete(self):
        # Place your code here:
        pass


app = init_job_controller(MyTokenListener)



import json
import os
import socket
import uuid
from os import listdir
from os.path import isdir, isfile, basename, join
from typing import Any
from bson import ObjectId
from pymongo import MongoClient
from pymongo.errors import OperationFailure, PyMongoError
from balticlsc.computation_module.baltic_lsc.data_handle import DataHandle
from balticlsc.computation_module.data_model.pins_configuration import Multiplicity
from balticlsc.computation_module.utils.logger import logger


def _download_single_file(document: Any, local_path: str) -> str:
    file_content = document['fileContent']
    file_path = os.path.join(local_path, document['fileName'])
    with open(file_path, mode='wb') as file:
        file.write(file_content)
    return file_path


class MongoDBHandle(DataHandle):

    def __init__(self, pin_name: str, pins_configuration: []):
        super().__init__(pin_name, pins_configuration)
        self.__connection_string = 'mongodb://' + self._pin_configuration.access_credential['User'] + ':' \
                                   + self._pin_configuration.access_credential['Password'] + '@' \
                                   + self._pin_configuration.access_credential['Host'] + ':' \
                                   + self._pin_configuration.access_credential['Port']
        self.__mongo_client = None
        self.__mongo_database = None
        self.__mongo_collection = None

    def download(self, handle: {}) -> str:
        if "input" != self._pin_configuration.pin_type:
            raise Exception('Download cannot be called for output pins')
        if 'Database' not in handle:
            raise ValueError('Incorrect DataHandle.')
        if 'Collection' not in handle:
            raise ValueError('Incorrect DataHandle.')
        collection_name = handle['Collection']
        self.__prepare(handle['Database'], collection_name)
        local_path = ''
        if Multiplicity.SINGLE == self._pin_configuration.data_multiplicity:
            if 'ObjectId' not in handle:
                raise ValueError('Incorrect DataHandle.')
            obj_id = handle['ObjectId']
            try:
                logger.debug('Downloading object with id: ' + obj_id)
                document = self.__mongo_collection.find_one({'_id': ObjectId(obj_id)})
                if document is not None:
                    local_path = _download_single_file(document, self._local_path)
                    logger.debug('Downloading object with id: ' + obj_id + ' successful.')
                else:
                    logger.debug('Can not find object with id ' + obj_id)
            except BaseException as e:
                logger.debug('Downloading object with id: ' + obj_id + ' failed.')
                self._clear_local()
                raise e
        elif Multiplicity.MULTIPLE == self._pin_configuration.data_multiplicity:
            try:
                logger.debug('Downloading all files from ' + collection_name)
                local_path = self._local_path + '/' + collection_name
                os.mkdir(local_path)
                documents = self.__mongo_collection.find()
                for doc in documents:
                    _download_single_file(doc, local_path)
                self._add_guids_to_file_names(local_path)
                logger.debug('Downloading all files from ' + collection_name + ' successful.')
            except BaseException as e:
                logger.debug('Downloading all files from ' + collection_name + ' failed.')
                self._clear_local()
                raise e
        return local_path

    def upload(self, local_path: str) -> {}:
        if "output" != self._pin_configuration.pin_type:
            raise Exception('Upload cannot be called for input pins')
        if not isfile(local_path) and not isdir(local_path):
            raise ValueError('Invalid path (' + local_path + ')')
        is_directory = isdir(local_path)
        if Multiplicity.MULTIPLE == self._pin_configuration.data_multiplicity and not is_directory:
            raise ValueError('Multiple data pin requires path pointing to a directory, not a file')
        if Multiplicity.SINGLE == self._pin_configuration.data_multiplicity and is_directory:
            raise ValueError('Single data pin requires path pointing to a file, not a directory')
        handle = {}
        try:
            database_name, collection_name = self.__prepare()
            if Multiplicity.SINGLE == self._pin_configuration.data_multiplicity:
                logger.debug('Uploading file from ' + local_path + 'to collection ' + collection_name)
                file_name = basename(local_path)
                with open(local_path, mode='rb') as file:
                    file_content = file.read()
                result = self.__mongo_collection.insert_one({'fileName': file_name,
                                                             'fileContent': file_content})
                handle = {'FileName': file_name, 'ObjectId': str(result.inserted_id), 'Database': database_name,
                          'Collection': collection_name}
                logger.debug('Uploading file from ' + local_path + ' to collection ' + collection_name +
                             ' successful.')
            elif Multiplicity.MULTIPLE == self._pin_configuration.data_multiplicity:
                logger.debug('Uploading directory from ' + local_path + ' to collection ' + collection_name)
                handle_list = []
                for f in (f for f in listdir(local_path) if isfile(join(local_path, f))):
                    file_name = basename(f)
                    with open(f, mode='rb') as file:
                        file_content = file.read()
                    result = self.__mongo_collection.insert_one({'fileName': file_name,
                                                                 'fileContent': file_content})
                    handle_list.append({'FileName': file_name, 'ObjectId': str(result.inserted_id)})
                handle = {'Files': json.dumps(handle_list), 'Database': database_name,
                          'Collection': collection_name}
                logger.Debug('Uploading directory from ' + local_path + 'to collection ' + collection_name +
                             ' successful.')
            return handle
        except BaseException as e:
            logger.Debug('Error: ' + str(e) + ' \n Uploading from ' + local_path + ' failed.')
            raise e
        finally:
            self._clear_local()

    def check_connection(self, handle: {}):
        host = self._pin_configuration.access_credential["Host"]
        port = int(self._pin_configuration.access_credential["Port"])
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        try:
            if 0 != sock.connect_ex((host, port)):
                logger.debug(f'Unable to reach {host}:{port}')
                return -1
        finally:
            sock.close()
        logger.debug(f'MongoDB connection string: {self.__connection_string}')

        try:
            self.__mongo_client = MongoClient(self.__connection_string)
            self.__mongo_client.list_databases()
        except OperationFailure:  # TODO check if authorization error (if needed)
            logger.debug('Unable to authenticate to MongoDB')
            return -2
        except PyMongoError as e:
            logger.debug('Error ' + str(e) + ' while trying to connect to MongoDB')
            return -1
        if 'input' == self._pin_configuration.pin_type and handle is not None:
            if 'Database' not in handle:
                raise ValueError('Incorrect DataHandle.')
            if 'Collection' not in handle:
                raise ValueError('Incorrect DataHandle.')
            database_name = handle['Database']
            collection_name = handle['Collection']
            if Multiplicity.SINGLE == self._pin_configuration.data_multiplicity and 'ObjectId' not in handle:
                raise ValueError('Incorrect DataHandle.')
            obj_id = handle['ObjectId']
            try:
                self.__mongo_database = self.__mongo_client[database_name]
                if self.__mongo_database is None:
                    logger.debug('No database ' + database_name)
                    return -3
                self.__mongo_collection = self.__mongo_database[collection_name]
                if self.__mongo_collection is None:
                    logger.debug('No collection ' + database_name)
                    return -3
                if Multiplicity.SINGLE == self._pin_configuration.data_multiplicity:
                    document = self.__mongo_collection.find_one({'_id': ObjectId(obj_id)})
                    if document is None:
                        logger.debug('No document with id ' + obj_id)
                        return -3
            except PyMongoError:
                logger.debug('Error while trying to ' +
                             ('access collection ' + collection_name if obj_id is None else 'get object ' + obj_id)
                             + ' from database ' + database_name +
                             ('' if obj_id is None else ' from collection ' + collection_name))
                return -3
        return 0

    def __prepare(self, database_name: str = None, collection_name: str = None) -> (str, str):
        if database_name is None:
            database_name = 'baltic_database_' + str(uuid.uuid4())[:8]
        if collection_name is None:
            collection_name = 'baltic_collection_' + str(uuid.uuid4())[:8]
        # TODO to reset or not to reset
        self.__mongo_client = MongoClient(self.__connection_string)
        self.__mongo_database = self.__mongo_client[database_name]
        self.__mongo_collection = self.__mongo_database[collection_name]
        return database_name, collection_name



import json
import os
from enum import Enum
from balticlsc.computation_module.utils.utils import camel_dict_to_snake_dict


class Multiplicity(Enum):
    SINGLE = 0
    MULTIPLE = 1


class PinConfiguration:
    pin_name: str
    pin_type: str
    is_required: str
    token_multiplicity: Multiplicity
    data_multiplicity: Multiplicity
    access_type: str
    access_credential: {}

    def __init__(self, pin_name: str, pin_type: str, is_required: str, token_multiplicity: Multiplicity,
                 data_multiplicity: Multiplicity, access_type: str, access_credential: {}):
        self.pin_name = pin_name
        self.pin_type = pin_type
        self.is_required = is_required
        self.token_multiplicity = token_multiplicity
        self.data_multiplicity = data_multiplicity
        self.access_type = access_type
        self.access_credential = access_credential


def get_pins_configuration() -> []:
    pins_config_path = os.getenv('SYS_PIN_CONFIG_FILE_PATH', '/app/module/configs/pins.json')
    with open(pins_config_path) as pins_config_file:
        try:
            pins_configuration = []
            for p in json.load(pins_config_file):
                try:
                    pin = PinConfiguration(
                        **{key: Multiplicity[value.upper()] if key in ('token_multiplicity',
                                                                       'data_multiplicity') else value
                           for key, value in camel_dict_to_snake_dict(p).items()
                           if key in PinConfiguration.__dict__['__annotations__']})
                except BaseException as lpe:
                    error_msg = 'Wrong config for pin - json:' + str(p) + ', error: ' + str(lpe)
                    raise ValueError(error_msg) from lpe
                pins_configuration.append(pin)
        except BaseException as lpe:
            error_msg = 'Error while loading pins config from ' + pins_config_path + ': ' + str(lpe)
            raise ValueError(error_msg) from lpe
    return pins_configuration



import abc
import os
import threading
from collections import defaultdict

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from balticlsc.computation_module.api_access.job_controller import JobThread

from balticlsc.computation_module.data_model.messages import Status, JobStatus, InputTokenMessage
from balticlsc.computation_module.data_model.pins_configuration import Multiplicity, PinConfiguration


class IJobRegistry(metaclass=abc.ABCMeta):

    @classmethod
    def __subclasshook__(cls, subclass):
        return (hasattr(subclass, 'get_pin_status') and
                callable(subclass.get_pin_status) and
                hasattr(subclass, 'get_pin_tokens') and
                callable(subclass.get_pin_tokens) and
                hasattr(subclass, 'get_pin_value') and
                callable(subclass.get_pin_value) and
                hasattr(subclass, 'get_pin_values') and
                callable(subclass.get_pin_values) and
                hasattr(subclass, 'get_pin_values_dim') and
                callable(subclass.get_pin_values_dim) and
                hasattr(subclass, 'get_progress') and
                callable(subclass.get_progress) and
                hasattr(subclass, 'get_variable') and
                callable(subclass.get_variable) and
                hasattr(subclass, 'set_progress') and
                callable(subclass.set_progress) and
                hasattr(subclass, 'set_status') and
                callable(subclass.set_status) and
                hasattr(subclass, 'set_variable') and
                callable(subclass.set_variable) and
                hasattr(subclass, 'get_environment_variable') and
                callable(subclass.get_environment_variable) or
                NotImplemented)

    @abc.abstractmethod
    def get_pin_status(self, pin_name: str) -> Status:
        pass

    @abc.abstractmethod
    def get_pin_tokens(self, pin_name: str) -> []:
        pass

    @abc.abstractmethod
    def get_pin_value(self, pin_name: str) -> str or None:
        pass

    @abc.abstractmethod
    def get_pin_values(self, pin_name: str) -> []:
        pass

    @abc.abstractmethod
    def get_pin_values_dim(self, pin_name: str) -> ([], []):
        pass

    @abc.abstractmethod
    def get_progress(self) -> int:
        pass

    @abc.abstractmethod
    def get_variable(self, name: str) -> object:
        pass

    @abc.abstractmethod
    def set_progress(self, progress: int):
        pass

    @abc.abstractmethod
    def set_status(self, status: Status):
        pass

    @abc.abstractmethod
    def set_variable(self, name: str, value: object):
        pass

    @abc.abstractmethod
    def get_environment_variable(self, name: str) -> str:
        pass


class JobRegistry(IJobRegistry):

    def __init__(self, pins_configuration: []):
        self.__pins = pins_configuration
        self.__tokens = defaultdict(list)
        self.__status = JobStatus(os.getenv('SYS_MODULE_INSTANCE_UID', 'module_uid'))
        self.__variables = {}
        self.__job_threads = []
        self.__semaphore = threading.Semaphore()

    def register_thread(self, thread: 'JobThread'):
        self.__semaphore.acquire()
        try:
            self.__job_threads.append(thread)
            return 0
        finally:
            self.__semaphore.release()

    def get_pin_status(self, pin_name: str) -> Status:
        self.__semaphore.acquire()
        try:
            if 0 == len(self.__tokens[pin_name]):
                return Status.IDLE
            if Multiplicity.SINGLE == self.get_pin_configuration(pin_name, False).token_multiplicity:
                return Status.COMPLETED
            final_token = next((t for t in self.__tokens[pin_name] if any(not s.is_final for s in t.token_seq_stack)),
                               None)
            if final_token is not None:
                max_count = 1
                for st in final_token.token_seq_stack:
                    max_count *= st.no + 1
                if len(self.__tokens[pin_name]) == max_count:
                    return Status.COMPLETED
            return Status.WORKING
        finally:
            self.__semaphore.release()

    def get_pin_tokens(self, pin_name: str) -> []:
        self.__semaphore.acquire()
        try:
            return self.__tokens[pin_name]
        finally:
            self.__semaphore.release()

    def get_pin_value(self, pin_name: str) -> str or None:
        (values, sizes) = self.get_pin_values_dim(pin_name)
        if values is None or 0 == len(values):
            return None
        if sizes is None and 1 == len(values):
            return values[0]
        raise Exception('Improper call - more than one token exists for the pin')

    def get_pin_values(self, pin_name: str) -> []:
        (values, sizes) = self.get_pin_values_dim(pin_name)
        if sizes is not None and 1 == len(sizes):
            return values
        raise Exception('Improper call - more than one dimension exists for the pin')

    def get_pin_values_dim(self, pin_name: str) -> ([], []):
        self.__semaphore.acquire()
        try:
            if 0 == len(self.__tokens[pin_name]):
                return None, None
            if Multiplicity.SINGLE == self.get_pin_configuration(pin_name, False).token_multiplicity:
                return [token.values for token in self.__tokens[pin_name]], None
            max_table_counts = [0] * len(next(self.__tokens[pin_name]).token_seq_stack)
            for m in self.__tokens[pin_name]:
                for i in range(len(m.token_seq_stack)):
                    if max_table_counts[i] < m.token_seq_stack[i].no:
                        max_table_counts[i] = m.token_seq_stack[i].no
            all_token_count = 1
            for i in max_table_counts:
                all_token_count *= i + 1
            result = [None] * all_token_count
            for m in self.__tokens[pin_name]:
                index = 0
                product = 1
                for i in range(len(m.token_seq_stack)):
                    index += max_table_counts[i] * product
                    product *= max_table_counts[i]
                result[index] = m.values
            return result, max_table_counts
        finally:
            self.__semaphore.release()

    def get_progress(self) -> int:
        self.__semaphore.acquire()
        try:
            return self.__status.job_progress
        finally:
            self.__semaphore.release()

    def get_job_status(self) -> JobStatus:
        self.__semaphore.acquire()
        try:
            return self.__status
        finally:
            self.__semaphore.release()

    def get_variable(self, name: str) -> object:
        self.__semaphore.acquire()
        try:
            return self.__variables[name]
        finally:
            self.__semaphore.release()

    def set_progress(self, progress: int):
        self.__semaphore.acquire()
        try:
            self.__status.job_progress = progress
        finally:
            self.__semaphore.release()

    def set_status(self, status: Status):
        self.__semaphore.acquire()
        try:
            self.__status.status = status
        finally:
            self.__semaphore.release()

    def set_variable(self, name: str, value: object):
        self.__semaphore.acquire()
        try:
            self.__variables[name] = object
        finally:
            self.__semaphore.release()

    def get_environment_variable(self, name: str) -> str:
        return os.getenv(name)

    def get_pin_configuration(self, pin_name: str, acquire: bool = True) -> PinConfiguration:
        if acquire:
            self.__semaphore.acquire()
        try:
            return next(p for p in self._JobRegistry__pins if pin_name == p.pin_name)
        finally:
            if acquire:
                self.__semaphore.release()

    def get_strong_pin_names(self) -> []:
        self.__semaphore.acquire()
        try:
            return list(p.pin_name for p in self._JobRegistry__pins if 'true' == p.is_required)
        finally:
            self.__semaphore.release()

    def get_base_msg_uid(self):
        self.__semaphore.acquire()
        try:
            return next(ltm for ltm in self._JobRegistry__tokens.values() if 0 != len(ltm))[0].msg_uid
        finally:
            self.__semaphore.release()

    def get_all_msg_uids(self):
        self.__semaphore.acquire()
        try:
            return list(it.msg_uid for pit in self._JobRegistry__tokens.values() for it in pit)
        finally:
            self.__semaphore.release()

    def clear_messages(self, msg_ids):
        for msg_id in msg_ids:
            tokens = next(ltm for ltm in self._JobRegistry__tokens.values() if any(msg_id == it.msg_uid for it in ltm))
            if tokens is not None:
                message = next(msg for msg in tokens if msg_id == msg.msg_uid)
                tokens.remove(message)

    def register_token(self, msg: InputTokenMessage):
        self.__tokens[msg.pin_name].append(msg)



from enum import Enum


class SeqToken:
    seq_uid: str
    no: int
    is_final: bool

    def __init__(self, seq_uid: str, no: int, is_final: bool):
        self.seq_uid = seq_uid
        self.no = no
        self.is_final = is_final


class InputTokenMessage:
    msg_uid: str
    pin_name: str
    values: str
    access_type: str
    token_seq_stack: []

    def __init__(self, msg_uid: str, pin_name: str, values: str, access_type: str, token_seq_stack: []):
        self.msg_uid = msg_uid
        self.pin_name = pin_name
        self.values = values
        self.access_type = access_type
        self.token_seq_stack = token_seq_stack


class Status(Enum):
    IDLE = 0
    WORKING = 1
    COMPLETED = 2
    FAILED = 3


class JobStatus:
    job_instance_uid: str
    status: Status
    job_progress: int

    def __init__(self, job_instance_uid: str, status: Status = Status.IDLE, job_progress: int = -1):
        self.job_instance_uid = job_instance_uid
        self.job_progress = job_progress
        self.status = status


class OutputTokenMessage:
    pin_name: str
    sender_uid: str
    values: str
    base_msg_uid: str
    is_final: bool

    def __init__(self, pin_name: str, sender_uid: str, values: str, base_msg_uid: str, is_final: bool):
        self.pin_name = pin_name
        self.sender_uid = sender_uid
        self.values = values
        self.base_msg_uid = base_msg_uid
        self.is_final = is_final


class TokensAck:
    sender_uid: str
    msg_uids: []
    note: str
    is_final: bool
    is_failed: bool

    def __init__(self, sender_uid: str, msg_ids: [], note: str, is_final: bool, is_failed: bool):
        self.sender_uid = sender_uid
        self.msg_uids = msg_ids
        self.note = note
        self.is_final = is_final
        self.is_failed = is_failed



import json
import os
from http import HTTPStatus
import requests
from balticlsc.computation_module.data_model.messages import OutputTokenMessage, TokensAck
from balticlsc.computation_module.utils.utils import snake_dict_to_camel_dict


class TokensProxy:

    def __init__(self):
        self.__sender_uid = os.getenv('SYS_MODULE_INSTANCE_UID', 'module_uid')
        self.__batch_manager_ack_url = os.getenv('SYS_BATCH_MANAGER_ACK_ENDPOINT', 'http://127.0.0.1:7000/ack')
        self.__batch_manager_token_url = os.getenv('SYS_BATCH_MANAGER_TOKEN_ENDPOINT', 'http://127.0.0.1:7000/token')

    def send_output_token(self, pin_name: str, values: str, msg_id: str, is_final: bool) -> HTTPStatus:
        output_token = OutputTokenMessage(pin_name, self._TokensProxy__sender_uid, values, msg_id, is_final)
        return HTTPStatus(requests.post(self._TokensProxy__batch_manager_token_url,
                                        data=json.dumps(snake_dict_to_camel_dict(output_token.__dict__)),
                                        headers={'content-type': 'application/json'}).status_code)

    def send_ack_token(self, msg_ids: [], is_final: bool, is_failed: bool = False, note: str = None) -> HTTPStatus:
        ack_token = TokensAck(self.__sender_uid, msg_ids, note, is_final, is_failed)
        return HTTPStatus(requests.post(self._TokensProxy__batch_manager_ack_url,
                                        data=json.dumps(snake_dict_to_camel_dict(ack_token.__dict__)),
                                        headers={'content-type': 'application/json'}).status_code)
